use std::collections::HashSet;
use std::path::PathBuf;

use clap::{value_parser, Arg, ArgAction, ArgMatches, Command};
use color_eyre::eyre::{self, Context};
use color_eyre::Result;
use itertools::Itertools;
use sdk::murmur::Murmur64;
use tokio::fs;
use tokio::io::AsyncWriteExt;
use tokio::time::Instant;

pub(crate) fn command_definition() -> Command {
    Command::new("brute-force-words")
        .about(
            "Given a list of words and a set of delimiters, iteratevily creates permutations \
                of growing length.\n\
                Delimiters are placed between every word in the result.\n\n\
                Example: \
                Given the words ['packages', 'boot'], the delimiters ['/', '_'] and a length of 2, the resulting \
                words will be\n\
                - packages\n\
                - boot\n\
                - packages/packages\n\
                - packages_packages\n\
                - packages/boot\n\
                - packages_boot\n\
                - boot/packages\n\
                - boot_packages\n\
                - boot/boot\n\
                - boot_boot",
        )
        .arg(
            Arg::new("delimiter")
                .help(
                    "The delimiters to put between the words. \
                        All permutations of this list will be tried for every string of words.\n\
                        Specify multiple times to set multiple values.\n\
                        Defaults to ['/', '_'].",
                )
                .short('d')
                .long("delimiter")
                .action(ArgAction::Append),
        )
        .arg(
            Arg::new("max-length")
                .help("The maximum number of words up to which to build strings.")
                .long("max")
                .long("max-length")
                .short('m')
                .default_value("5")
                .value_parser(value_parser!(usize)),
        )
        .arg(
            Arg::new("continue")
                .help("Can be used to continue a previous operation where it stopped. Word list and delimiters must match.")
                .short('c')
                .long("continue")
        )
        .arg(
            Arg::new("words")
                .help("Path to a file containing words line by line.")
                .required(true)
                .value_parser(value_parser!(PathBuf)),
        )
        .arg(
            Arg::new("hashes")
                .help(
                    "Path to a file containing the hashes to attempt to brute force. \
                        Hashes are expected in hexadecimal notiation. \
                        Only 64-bit hashes are supported."
                )
                .required(true)
                .value_parser(value_parser!(PathBuf)),
        )
}

#[tracing::instrument(skip_all)]
#[allow(clippy::mut_range_bound)]
pub(crate) async fn run(_ctx: sdk::Context, matches: &ArgMatches) -> Result<()> {
    let max_length: usize = matches
        .get_one::<usize>("max-length")
        .copied()
        .expect("parameter has default");

    let words: Vec<String> = {
        let path = matches
            .get_one::<PathBuf>("words")
            .expect("missing required parameter");

        let file = fs::read_to_string(&path)
            .await
            .wrap_err_with(|| format!("Failed to read file '{}'", path.display()))?;

        file.lines().map(str::to_string).collect()
    };

    if words.is_empty() {
        eyre::bail!("Word list must not be empty");
    }

    let hashes = {
        let path = matches
            .get_one::<PathBuf>("hashes")
            .expect("missing required argument");
        let content = fs::read_to_string(&path)
            .await
            .wrap_err_with(|| format!("Failed to read file '{}'", path.display()))?;

        let hashes: Result<HashSet<_>, _> = content
            .lines()
            .map(|s| u64::from_str_radix(s, 16).map(Murmur64::from))
            .collect();

        let hashes = hashes?;

        tracing::trace!("{:?}", hashes);

        hashes
    };

    let mut delimiters: Vec<String> = matches
        .get_many::<String>("delimiter")
        .unwrap_or_default()
        .cloned()
        .collect();

    if delimiters.is_empty() {
        delimiters.push(String::from("/"));
        delimiters.push(String::from("_"));
    }

    let delimiters_len = delimiters.len();

    let prefixes = [
        "",
        "content/characters/",
        "content/debug/",
        "content/decals/",
        "content/environment/",
        "content/fx/",
        "content/gizmos/",
        "content/items/",
        "content/levels/",
        "content/liquid_area/",
        "content/localization/",
        "content/materials/",
        "content/minion_impact_assets/",
        "content/pickups/",
        "content/shading_environments/",
        "content/textures/",
        "content/ui/",
        "content/videos/",
        "content/vo/",
        "content/volume_types/",
        "content/weapons/",
        "packages/boot_assets/",
        "packages/content/",
        "packages/game_scripts/",
        "packages/strings/",
        "packages/ui/",
        "wwise/events/",
        "wwise/packages/",
        "wwise/world_sound_fx/",
    ];

    let word_count = words.len();
    tracing::info!("{} words to try", word_count);

    // To be able to easily combine the permutations of words and delimiters,
    // we turn the latter into a pre-defined list of all permutations of delimiters
    // that are possible at the given amount of words.
    // Combining `Iterator::cycle` with `Itertools::permutations` works, but
    // with a high `max_length`, it runs OOM.
    // So we basically have to implement a smaller version of the iterative algorithm we use later on
    // to build permutations of the actual words.
    let delimiter_lists = {
        let mut indices = vec![0; max_length - 1];
        let mut list = Vec::new();

        for _ in 0..delimiters_len.pow(max_length as u32 - 1) {
            list.push(indices.iter().map(|i| &delimiters[*i]).collect::<Vec<_>>());

            for v in indices.iter_mut() {
                if *v >= delimiters_len - 1 {
                    *v = 0;
                    break;
                } else {
                    *v += 1;
                }
            }
        }

        list
    };

    tracing::debug!("{:?}", delimiter_lists);

    let mut indices = if let Some(cont) = matches.get_one::<String>("continue").cloned() {
        let mut splits = vec![cont.clone()];

        for delim in delimiters.iter() {
            splits = splits
                .iter()
                .flat_map(|s| s.split(delim))
                .map(|s| s.to_string())
                .collect();
        }

        let indices = splits
            .into_iter()
            .map(|s| {
                words
                    .iter()
                    .enumerate()
                    .find(|(_, v)| s == **v)
                    .map(|(i, _)| i)
                    .ok_or_else(|| eyre::eyre!("'{}' is not in the word list", s))
            })
            .collect::<Result<_>>()?;

        tracing::info!("Continuing from '{}' -> '{:?}'", cont, &indices);

        indices
    } else {
        vec![0]
    };
    let mut indices_len = indices.len();
    let mut sequence = indices
        .iter()
        .map(|index| words[*index].as_str())
        .collect::<Vec<_>>();

    // Prevent re-allocation by reserving as much as we need upfront
    indices.reserve(max_length);
    sequence.reserve(max_length);

    let mut count: usize = 0;
    let mut found: usize = 0;
    let mut start = Instant::now();

    // let mut writer = BufWriter::new(tokio::io::stdout());
    let mut writer = tokio::io::stdout();
    let mut buf = Vec::with_capacity(1024);

    const LINE_FEED: u8 = 0x0A;
    const UNDERSCORE: u8 = 0x5F;
    const ZERO: u8 = 0x30;

    'outer: loop {
        // We only want delimiters between words, so we keep that iterator shorter by
        // one.
        let delimiter_count = sequence.len() as u32 - 1;

        for prefix in prefixes.iter().map(|p| p.as_bytes()) {
            buf.clear();

            // We can keep the prefix at the front of the buffer and only
            // replace the parts after that.
            let prefix_len = prefix.len();
            buf.extend_from_slice(prefix);

            for delims in delimiter_lists
                .iter()
                .take(delimiters_len.pow(delimiter_count))
            {
                buf.truncate(prefix_len);

                let delims = delims
                    .iter()
                    .map(|s| s.as_str())
                    .take(delimiter_count as usize);
                sequence
                    .iter()
                    .copied()
                    .interleave(delims.clone())
                    .for_each(|word| buf.extend_from_slice(word.as_bytes()));

                count += 1;

                let hash = Murmur64::hash(&buf);
                if hashes.contains(&hash) {
                    found += 1;
                    buf.push(LINE_FEED);
                    writer.write_all(&buf).await?;
                } else {
                    let word_len = buf.len();

                    // If the regular word itself didn't match, we check
                    // for numbered suffixes.
                    // For now, we only check up to `09` to avoid more complex logic
                    // writing into the buffer.
                    // Packages that contain files with higher numbers than this
                    // should hopefully become easier to spot once a good number of
                    // hashes is found.
                    for i in 1..=9 {
                        buf.truncate(word_len);
                        buf.push(UNDERSCORE);
                        buf.push(ZERO);
                        buf.push(ZERO + i);

                        count += 1;

                        let hash = Murmur64::hash(&buf);
                        if hashes.contains(&hash) {
                            found += 1;
                            buf.push(LINE_FEED);
                            writer.write_all(&buf).await?;
                        } else {
                            break;
                        }
                    }
                }
            }
        }

        let dur = Instant::now() - start;
        if dur.as_secs() >= 1 {
            let hashes_len = hashes.len();
            let s = String::from_utf8_lossy(&buf);
            // The last prefix in the set is the one that will stay in the buffer
            // when we're about to print here.
            // So we strip that, to show just the generated part.
            // We also restrict the length to stay on a single line.
            let prefix_len = prefixes[28].len();
            let s = s[prefix_len..std::cmp::min(s.len(), prefix_len + 60)]
                .trim_end()
                .to_string();
            // Don't care when it finishes, don't care if it fails.
            tokio::spawn(async move {
                let _ = tokio::io::stderr()
                    .write_all(
                        format!(
                            "\r{:8} hashes per second | {:6}/{} found | {:<60}",
                            count, found, hashes_len, s
                        )
                        .as_bytes(),
                    )
                    .await;
            });

            start = Instant::now();
            count = 0;
        }

        for i in 0..indices_len {
            let index = indices.get_mut(i).unwrap();
            let word = sequence.get_mut(i).unwrap();

            if *index >= word_count - 1 {
                *index = 0;
                *word = words[*index].as_str();

                if indices.get(i + 1).is_none() {
                    indices_len += 1;

                    if indices_len > max_length {
                        break 'outer;
                    }

                    indices.push(0);
                    sequence.push(words[0].as_str());

                    break;
                }
            } else {
                *index += 1;
                *word = words[*index].as_str();
                break;
            }
        }
    }

    Ok(())
}
